﻿using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SerializationTest.Model;

namespace SerializationTest
{
    [TestClass]
    public class BinaryFormatterTest
    {
        #region 不加Serializable就不能序列化，而且要求所有基类和成员类都要加。
        [TestMethod]
        public void AttributeMissingOnPublicClassTest()
        {
            try
            {
                var obj = new PublicClassWithoutAttribute()
                {
                    Name = "Name",
                    Value = "Value",
                    Age = 1
                };
                SerializeAndPrintAndDeserialize(obj);
                Assert.Fail();
            }
            catch (SerializationException e)
            {
                Trace.TraceInformation(e.ToString());
            }

            try
            {
                var obj = SerializeAndPrintAndDeserialize(new SerializableBookInfo());
                Assert.Fail();
            }
            catch (SerializationException e)
            {
                Trace.TraceInformation(e.ToString());
            }
        } 
        #endregion

        #region 默认序列化所有字段
        [TestMethod]
        public void BasicObjectFormattTest()
        {
            var obj = new Staff() { Name = "Hugo", Age = 30, Salary = 1000 };
            var obj2 = SerializeAndPrintAndDeserialize(obj) as Staff;
            Assert.IsNotNull(obj2);
            Assert.AreEqual(obj.Name, obj2.Name);
            Assert.AreEqual(obj.Age, obj2.Age);
            Assert.IsFalse(obj.RicherThan(obj2));
            Assert.IsFalse(obj2.RicherThan(obj));
        } 
        #endregion

        #region 支持序列化包含循环引用的对象，并能正确还原引用关系。
        [TestMethod]
        public void SelfCircleSerializeTest()
        {
            var obj = new LinkNode(new LinkNode());
            var obj2 = SerializeAndPrintAndDeserialize(obj) as LinkNode;
            Assert.IsNotNull(obj2);
            Assert.IsNotNull(obj2.Child);
            CollectionAssert.AllItemsAreNotNull(obj2.Child.Select(c => c.Parent).ToList());
            Assert.IsTrue(obj2.Child.All(c => c.Parent == obj2));
        } 
        #endregion

        #region 对于简单对象而言，BinaryFormatter和DataContractSerializer的性能也在一个数量级上
        [TestMethod]
        public void SerializationPerformanceTest()
        {
            var obj = new LinkNode(new LinkNode());
            var times = 10000;
            var i = 0;
            SerializeAndPrintAndDeserialize(obj, false);
            var watch = Stopwatch.StartNew();
            while (i++ < times)
            {
                var sample = SerializeAndPrintAndDeserialize(obj, false);

            }
            var binaryTime = watch.ElapsedMilliseconds;
            DataContractSerializerTest.SerializeAndPrintAndDeserialize(obj, false);
            watch.Restart();
            i = 0;
            while (i++ < times)
            {
                var result = DataContractSerializerTest.SerializeAndPrintAndDeserialize(obj, false);
            }
            var xmlTime = watch.ElapsedMilliseconds;

            Trace.TraceInformation(String.Format("BinaryFormatter用时：{0}", binaryTime));
            Trace.TraceInformation(String.Format("DataContractSerializer用时：{0}", xmlTime));


            Assert.IsTrue(binaryTime * 2 > xmlTime);
            Assert.IsTrue(binaryTime < xmlTime * 2);
        } 
        #endregion

        #region 对于包含多个重复引用的对象而言，BinaryFormatter和DataContractSerializer的性能也在一个数量级上
        [TestMethod]
        public void SerializationMultipleReferencePerformanceTest()
        {
            var innerNode = new LinkNode(new LinkNode());
            var obj = new LinkNode();

            // Node中有一个对象的20次引用。
            int count = 20;
            while (count-- > 0)
                obj.Child.Add(innerNode);

            var times = 10000;
            var i = 0;
            SerializeAndPrintAndDeserialize(obj, false);
            var watch = Stopwatch.StartNew();
            while (i++ < times)
            {
                var sample = SerializeAndPrintAndDeserialize(obj, false);

            }
            var binaryTime = watch.ElapsedMilliseconds;
            DataContractSerializerTest.SerializeAndPrintAndDeserialize(obj, false);
            watch.Restart();
            i = 0;
            while (i++ < times)
            {
                var result = DataContractSerializerTest.SerializeAndPrintAndDeserialize(obj, false);
            }
            var xmlTime = watch.ElapsedMilliseconds;

            Trace.TraceInformation(String.Format("BinaryFormatter用时：{0}", binaryTime));
            Trace.TraceInformation(String.Format("DataContractSerializer用时：{0}", xmlTime));


            Assert.IsTrue(binaryTime * 2 > xmlTime);
            Assert.IsTrue(binaryTime < xmlTime * 2);
        }
        #endregion

        #region 对于简单对象BinaryFormatter和IXmlSerializer不相上下，但DataContractSerializer第一次会很慢。
        [TestMethod]
        public void CustomSerializationPerformanceTest()
        {
            var times = 1000;
            var obj = new CustomXmlSerialize() { Name = "Hugo", Address = "N/A", Value = "" };
            var i = 0;
            SerializeAndPrintAndDeserialize(obj, false);
            var watch = Stopwatch.StartNew();
            while (i++ < times)
            {
                var sample = SerializeAndPrintAndDeserialize(obj, false);

            }
            var binaryTime = watch.ElapsedMilliseconds;
            DataContractSerializerTest.SerializeAndPrintAndDeserialize(obj, false);
            watch.Restart();
            i = 0;
            while (i++ < times)
            {
                var result = DataContractSerializerTest.SerializeAndPrintAndDeserialize(obj, false);
            }
            var xmlTime = watch.ElapsedMilliseconds;

            Trace.TraceInformation(String.Format("BinaryFormatter用时：{0}", binaryTime));
            Trace.TraceInformation(String.Format("DataContractSerializer用时：{0}", xmlTime));

            Assert.IsTrue(binaryTime * 2 > xmlTime);
            Assert.IsTrue(binaryTime < xmlTime * 2);
        }  
        #endregion

        #region 支持IDeserializationCallback状态回调和OnSerializing等Attribute
        [TestMethod]
        public void CallbacksTest()
        {
            var obj = new Staff();
            var n = SerializeAndPrintAndDeserialize(obj) as Staff;
            Assert.AreEqual("Deserialized", n.Name);

            var call = new SerializableClassWithCallbacks();
            var call2 = SerializeAndPrintAndDeserialize(call) as SerializableClassWithCallbacks;

            Assert.IsNotNull(call2.Status);
            Assert.AreEqual("Serializing", call2.Status);
            Assert.AreEqual("Serialized", call.Status);
        }
        #endregion

        private static object SerializeAndPrintAndDeserialize(object instance, bool traceInfo = true)
        {
            var ms = new MemoryStream();
            var serializer = new BinaryFormatter();
            serializer.Serialize(ms, instance);
            ms.Seek(0, SeekOrigin.Begin);
            if (traceInfo)
            {
                using (var sr = new StreamReader(ms))
                {
                    var result = sr.ReadToEnd();
                    Trace.TraceInformation(String.Format("序列化结果的长度：{0} bytes", ms.Length));
                    Trace.TraceInformation(String.Format("序列化结果：{0}", result));
                    ms.Seek(0, SeekOrigin.Begin);
                    var i = serializer.Deserialize(ms);
                    Assert.IsNotNull(i);
                    Assert.IsInstanceOfType(i, instance.GetType());

                    return i;
                }
            }
            else
            {
                // 哪怕在转换间加入了Base64编解码的过程，也并不会减慢速度的。
                // return serializer.Deserialize(new MemoryStream(Convert.FromBase64String(Convert.ToBase64String(ms.ReadBytes()))));
                return serializer.Deserialize(ms);
            }
        }
    }
}
