﻿using System;
using System.Diagnostics;
using System.IO;
using System.Runtime.Serialization;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SerializationTest.Model;

namespace SerializationTest
{
    [TestClass]
    public class DataContractSerializerTest
    {
        #region Public类可以没有DataContract Attribute，默认序列化所有Public属性和字段。
        [TestMethod]
        public void DataContractAttributeMissingOnPublicClassTest()
        {
            var str = SerializeAndPrintAndDeserialize(new PublicClassWithoutAttribute());
            Assert.IsTrue(str.Contains("Name"));
            Assert.IsTrue(str.Contains("Value"));
            Assert.IsFalse(str.Contains("Age"));
        }
        #endregion

        #region DataContractSerializer 识别 Serializable Attribute，并序列化所有字段，不序列化属性，优先级低于DataContract
        [TestMethod]
        public void SerializablePublicClassTest()
        {
            var str = SerializeAndPrintAndDeserialize(new PublicClassWithSerializableAttribute());
            Assert.IsTrue(str.Contains("Name"));
            Assert.IsTrue(str.Contains("Value"));
            Assert.IsTrue(str.Contains("Age"));
            Assert.IsTrue(str.Contains("BirthDay"));
        } 
        #endregion

        #region 加上DataContract Attibute，可以序列化Private属性和字段
        [TestMethod]
        public void SerializeInternalMemberTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassWithInternalPropertyToSerialize());
            Assert.IsTrue(str.Contains("Age"));
        } 
        #endregion

        #region Internal类一定要有DataContract Attribute，才能被序列化
        [TestMethod]
        public void DataContractAttributeMissingOnInternalClassTest()
        {
            try
            {
                SerializeAndPrintAndDeserialize(new InternalClassWithoutContractAttribute());
                Assert.Fail();
            }
            catch (InvalidDataContractException e)
            {
                Trace.TraceInformation(e.ToString());
            }
        }
        #endregion

        #region 有DataContract的类，仅带有DataMember属性的成员才被序列化
        [TestMethod]
        public void DataContractWithoutMemberOnPublicClassTest()
        {
            var str = SerializeAndPrintAndDeserialize(new PublicClassWithContractAttribute());
            Assert.IsFalse(str.Contains("Name"));
            Assert.IsFalse(str.Contains("Value"));
        }
        #endregion

        #region 只读属性必须带上DataMember和DataContract才能被序列化
        [TestMethod]
        public void SerializeReadonlyProeprtyTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassWithReadOnlyPropertyToSerialize());
            Assert.IsTrue(str.Contains("Name"));
            Assert.IsFalse(str.Contains("Age"));
        } 
        #endregion

        #region DataContractSerializer 与 ISerializable接口 不兼容，无论是否标记了DataContract.
        [TestMethod]
        public void CompatibleWithISerializableTest()
        {
            try
            {
                SerializeAndPrintAndDeserialize(new ClassImplementsISerializable());
                Assert.Fail();
            }
            catch (InvalidDataContractException e)
            {
                Trace.TraceInformation(e.ToString());
            }
        } 
        #endregion

        #region DataContractSerializer支持IXmlSerializer接口，但不能标记DataContract.
        [TestMethod]
        public void CompatibleWithIXmlSerializableTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassImplementsIXmlSerializable());
            Assert.IsTrue(str.Contains("name="));
        } 
        #endregion

        #region 支持序列化Dictionary, 推论：与XmlSerializer不兼容。
        [TestMethod]
        public void DictionarySerializationTest()
        {
            SerializeAndPrintAndDeserialize(new ClassWithDictionary());
        }
        #endregion

        #region 属性值为null默认会被序列化，有EmitDefaultValue属性可以控制这个行为
        [TestMethod]
        public void SerializeNullTest()
        {
            var obj = new PublicClassWithoutAttribute();
            var str = SerializeAndPrintAndDeserialize(obj);
            Assert.IsNull(obj.Name);
            Assert.IsNull(obj.Value);
            Assert.IsTrue(str.Contains("Name") && str.Contains("Value"));
        }
        #endregion

        #region 加DataContract参数后，支持循环引用
        [TestMethod]
        public void SelfCircleSerializeTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassWithSelfCircle());
            Assert.IsTrue(str.Contains("Ref"));
            Assert.IsTrue(str.Contains("Child"));
            Assert.IsTrue(str.Contains("Parent"));
        } 
        #endregion

        #region 支持ElementName自定义
        [TestMethod]
        public void CustomElementName()
        {
            var str = SerializeAndPrintAndDeserialize(new ContractClassWithCustomXmlName());
            Assert.IsTrue(str.Contains("name"));
        } 
        #endregion

        #region 不支持XmlAttribute

        [TestMethod]
        public void NotSupportXmlAttributeTest()
        {
            // Xml开头的一序列Attribute本来就是给XmlSerializer用的。
            var str = SerializeAndPrintAndDeserialize(new ClassWithXmlAttribute());
            Assert.IsFalse(str.Contains("Name="));
        }

        #endregion

        #region 支持状态回调
        [TestMethod]
        public void CallbacksTest()
        {
            var obj = new ClassWithCallBacks();
            var str = SerializeAndPrintAndDeserialize(obj);
            Assert.IsTrue(str.Contains("Serializing"));
            Assert.AreEqual(obj.Status, "Serialized");
        } 
        #endregion

        #region 在DataMember中指定的顺序，会同时影响序列化和反序列化。（这点并不符合XML声明式标记语言的定义）
        /// <summary>
        /// 
        /// </summary>
        [TestMethod]
        public void PropertySerquencTest()
        {
            var xml = "<class><name>name</name><value>value</value></class>";
            var obj = ReadObject<ClassSpecifySequence>(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            Assert.AreEqual("name", obj.Name);
            Assert.AreEqual("value", obj.Value);

            xml = "<class><value>value</value><name>name</name></class>";
            obj = ReadObject<ClassSpecifySequence>(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            Assert.AreEqual("value", obj.Value);
            Assert.AreEqual(null, obj.Name);
        } 
        #endregion

        #region 在DataMember中不指定的顺序，也不能乱序读。
        /// <summary>
        /// 
        /// </summary>
        [TestMethod]
        public void PropertyNotSerquencTest()
        {
            var xml = "<class><name>name</name><value>value</value></class>";
            var obj = ReadObject<ClassNotSpecifySequence>(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            Assert.AreEqual("name", obj.Name);
            Assert.AreEqual("value", obj.Value);

            xml = "<class><value>value</value><name>name</name></class>";
            obj = ReadObject<ClassNotSpecifySequence>(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            Assert.AreEqual("value", obj.Value);
            // 序列不对就读不出来了
            // 所以DataContractSerializer只适用于严格规范化定义与生成的通信数据
            // 而无法读取网络上松散的Xml，如RSS和ATOM等。
            // 但这能带来多少解析的性能的提升？
            Assert.AreEqual(null, obj.Name);
        }
        #endregion

        #region 默认将枚举值序列化成字符串，可以使用EnumMember + DataContract修改
        [TestMethod]
        public void EnumValueSerializationTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassWithEnumValue());

            Assert.IsTrue(str.Contains("man"));
        } 
        #endregion

        private static T ReadObject<T>(Stream stream)
        {
            var serializer = new DataContractSerializer(typeof(T));

            return (T)serializer.ReadObject(stream);
        }

        public static string SerializeAndPrintAndDeserialize(object instance, bool traceInfo = true)
        {
            var ms = new MemoryStream();
            var serializer = new DataContractSerializer(instance.GetType());
            serializer.WriteObject(ms, instance);
            ms.Seek(0, SeekOrigin.Begin);
            if (traceInfo)
            {
                using (var sr = new StreamReader(ms))
                {
                    var result = sr.ReadToEnd();
                    Trace.TraceInformation(String.Format("序列化结果的大小为：{0} bytes", ms.Length));
                    Trace.TraceInformation(String.Format("序列化结果为：{0}", result));
                    ms.Seek(0, SeekOrigin.Begin);
                    var i = serializer.ReadObject(ms);
                    Assert.IsNotNull(i);
                    Assert.IsInstanceOfType(i, instance.GetType());

                    return result;
                }
            }
            else
            {
                return serializer.ReadObject(ms).ToString();
            }
        }
    }
}
