﻿using System;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Xml.Serialization;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SerializationTest.Model;

namespace SerializationTest
{
    [TestClass]
    public class XmlSerializerTest
    {
        #region 不加任何Attribute，XmlSerializer默认序列化Public的属性和字段。
        [TestMethod]
        public void AttributeMissingOnPublicClassTest()
        {
            var obj = new PublicClassWithoutAttribute()
            {
                Name = "Name",
                Value = "Value",
                Age = 1
            };
            var str = SerializeAndPrintAndDeserialize(obj);
            Assert.IsTrue(str.Contains("Name"));
            Assert.IsTrue(str.Contains("Value"));
            Assert.IsFalse(str.Contains("Age"));
        }
        #endregion

        #region 加上XmlElement Attibute，也不能序列化Private属性和字段

        /// <summary>
        /// 应该有一个AttributeScopeAttribute，控制一个Attribute可以放在Public,Internal,还是private上。
        /// 这些可以产生编译时错误，而不是到运行时才出现非预期的错误。
        /// </summary>
        [TestMethod]
        public void SerializeInternalMemberTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassWithInternalPropertyToSerialize());
            Assert.IsFalse(str.Contains("Age"));
        }
        #endregion

        #region 带不带Serializable结果是一样的。
        [TestMethod]
        public void WithAllEqualToWithoutAllTest()
        {
            var allstr = SerializeAndPrintAndDeserialize(new SerializableBookInfo());
            var nonstr = SerializeAndPrintAndDeserialize(new BookInfo());

            Assert.AreEqual(allstr, nonstr);
        }
        #endregion

        #region Null值不会被序列化

        [TestMethod]
        public void SerializeNullTest()
        {
            var obj = new PublicClassWithoutAttribute();
            var str = SerializeAndPrintAndDeserialize(obj);
            Assert.IsNull(obj.Name);
            Assert.IsNull(obj.Value);
            Assert.IsFalse(str.Contains("Name") || str.Contains("Value"));
        }

        #endregion

        #region 不支持序列化循环引用的对象
        [TestMethod]
        public void SelfCircleSerializeTest()
        {
            try
            {
                var str = SerializeAndPrintAndDeserialize(new ClassWithSelfCircle());
                Assert.Fail();
            }
            catch (InvalidOperationException e)
            {
                Trace.TraceInformation(e.ToString());
            }
        }
        #endregion

        #region DataContractSerializer 与 ISerializable接口 兼容，但并没有实际用到.
        [TestMethod]
        public void CompatibleWithISerializableTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassImplementsISerializable());

            Assert.IsNotNull(str);
            Assert.IsTrue(str.Contains(typeof(ClassImplementsISerializable).Name));
        }
        #endregion

        #region DataContractSerializer支持IXmlSerializer接口，但是会使Xml系列Attributes失效
        [TestMethod]
        public void CompatibleWithIXmlSerializableTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassImplementsIXmlSerializable());
            Assert.IsTrue(str.Contains("name="));
            Assert.IsFalse(str.Contains("Name"));
        }
        #endregion

        #region 不支持序列化Dictionary, 推论：与DataContractSerializer不兼容。
        [TestMethod]
        public void DictionarySerializationTest()
        {
            try
            {
                SerializeAndPrintAndDeserialize(new ClassWithDictionary());
                Assert.Fail();
            }
            catch (InvalidOperationException e)
            {
                Assert.IsInstanceOfType(e.InnerException, typeof(NotSupportedException));
            }
        }
        #endregion

        #region 不支持状态回调
        [TestMethod]
        public void CallbacksTest()
        {
            var obj = new ClassWithCallBacks();
            var str = SerializeAndPrintAndDeserialize(obj);
            Assert.IsFalse(str.Contains("Serializing"));
            Assert.AreEqual(obj.Status, null);
        }
        #endregion

        #region 支持XmlAttributes以精确控制序列化出来的结果

        [TestMethod]
        public void NotSupportXmlAttributeTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassWithXmlAttribute());
            Assert.IsTrue(str.Contains("Name="));
            str = SerializeAndPrintAndDeserialize(new BookInfo());
            // 自己看Xml
        }

        #endregion

        #region 不支持只读属性的序列化
        [TestMethod]
        public void SerializeReadonlyProeprtyTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassWithReadOnlyPropertyToSerialize());
            Assert.IsFalse(str.Contains("Name"));
            Assert.IsFalse(str.Contains("Age"));
        }
        #endregion

        #region 默认将枚举值序列化成字符串，可以使用XmlEnum修改
        [TestMethod]
        public void EnumValueSerializationTest()
        {
            var str = SerializeAndPrintAndDeserialize(new ClassWithEnumValue());

            Assert.IsTrue(str.Contains("man"));
        }
        #endregion

        #region 在XmlElement中指定的顺序，会同时影响序列化和反序列化。（这点并不符合XML声明式标记语言的定义）
        /// <summary>
        /// 
        /// </summary>
        [TestMethod]
        public void PropertySerquencTest()
        {
            var xml = "<class><name>name</name><value>value</value></class>";
            var obj = ReadObject<ClassSpecifySequence>(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            Assert.AreEqual("name", obj.Name);
            Assert.AreEqual("value", obj.Value);

            xml = "<class><value>value</value><name>name</name></class>";
            obj = ReadObject<ClassSpecifySequence>(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            Assert.AreEqual("value", obj.Value);
            Assert.AreEqual(null, obj.Name);
        }
        #endregion

        #region 在XmlElement中不指定的顺序，可以乱序读。
        /// <summary>
        /// 
        /// </summary>
        [TestMethod]
        public void PropertyNotSerquencTest()
        {
            var xml = "<class><name>name</name><value>value</value></class>";
            var obj = ReadObject<ClassNotSpecifySequence>(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            Assert.AreEqual("name", obj.Name);
            Assert.AreEqual("value", obj.Value);

            xml = "<class><value>value</value><name>name</name></class>";
            obj = ReadObject<ClassNotSpecifySequence>(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            Assert.AreEqual("value", obj.Value);
            Assert.AreEqual("name", obj.Name);
        }
        #endregion

        private static T ReadObject<T>(Stream stream)
        {
            var serializer = new XmlSerializer(typeof(T));

            return (T)serializer.Deserialize(stream);
        }

        public static string SerializeAndPrintAndDeserialize(object instance, bool tracing = true)
        {
            var ms = new MemoryStream();
            var serializer = new XmlSerializer(instance.GetType());
            serializer.Serialize(ms, instance);
            ms.Seek(0, SeekOrigin.Begin);
            if (tracing)
            {
                using (var sr = new StreamReader(ms))
                {
                    var result = sr.ReadToEnd();
                    Trace.TraceInformation(String.Format("序列化结果的大小为：{0} bytes", ms.Length));
                    Trace.TraceInformation(String.Format("序列化结果为：{0}", result));
                    ms.Seek(0, SeekOrigin.Begin);
                    var i = serializer.Deserialize(ms);
                    Assert.IsNotNull(i);
                    Assert.IsInstanceOfType(i, instance.GetType());

                    return result;
                }
            }
            else
            {
                return serializer.Deserialize(ms).ToString();
            }
        }
    }
}
